// PROFILE.C
//
// Profile management for X36 driver.
// 


// #define DBG 1

#include <WDM.H>
#include <filter.h>

// TEMP VARIABLE FOR OLD -> NEW INTERFACE CONVERSION
unsigned char CurrentCommandNum;
//



PSCRIPT ParseScript(IN PVOID Data, IN ULONG Length) // Converts a reg key into a script.
{
	PSCRIPT Head = NULL;
	PSCRIPT Current,Previous = NULL;
	PVOID Pos;
	unsigned long i=0;
	unsigned long j=0;
	
	DbgPrn(("ParseScript:  Length: %d\n",Length));	
	Pos=Data;
	Head = (PSCRIPT) ExAllocatePool(NonPagedPool, sizeof(SCRIPT));

	Current = Head;
	for (i=0; i+3<Length; i+=4)
	{
		DbgPrn(("ParseScript: count %d of length %d\n",i,Length));
		if (Current == NULL)
			i=Length;
		else
		{	
			RtlZeroMemory(Current,sizeof(SCRIPT));
			RtlCopyMemory(Current,Pos,2);
			Pos=(unsigned char*)Pos+2;
			//?
			j = (unsigned long) *(unsigned short*)((unsigned char*)Pos);
			Current->Delay = j * 10000;
			Pos=(unsigned char*)Pos+2;
			DbgPrn(("ParseScript:  Adding delay: %d   Offset: %d    Op: %x\n",Current->Delay, Current->Offset, Current->Op));	
			Current->Next=ExAllocatePool(NonPagedPool, sizeof(SCRIPT));
			Previous=Current;
			Current=Current->Next;
		}
	}
	if(Current != NULL)
	{
		ExFreePool(Current);
		if (Previous)
			Previous->Next = NULL;
		else
			Head=NULL;
	}
	
	return Head;
}	
	


 NTSTATUS AddEntry ( IN PWSTR ValueName, IN ULONG ValueType, IN PVOID ValueData, IN ULONG ValueLength, IN PVOID Context, IN PVOID EntryContext )
{
	unsigned short devicetype, number, command, switch1, switch2, shift, direction;
	struct DEVICE_EXTENSION *devExt = (struct DEVICE_EXTENSION *) Context;
	UNICODE_STRING  Name;
	unsigned long code;
	
	DbgPrn(("AddEntry:  Name: %S   Data: %X\n",ValueName,*(unsigned int *)ValueData));
	
	if ((ValueType != REG_BINARY) || (ValueName[0]==0))
		return STATUS_SUCCESS;
	
	RtlInitUnicodeString(&Name,ValueName);
	if (RtlUnicodeStringToAnsiSize(&Name)<7) // If it is /not/ a button control code. 
	{
		// Might be a script code.
		if (RtlUnicodeStringToInteger( &Name, 10, &code) == STATUS_SUCCESS)
		{
			if ((code > 0)&&(code<256))
			{
				devExt->Commands[code]=ParseScript(ValueData, ValueLength);
			}
		}
			
		return STATUS_SUCCESS;
	}
	else // Is a start command.
	{
		
		// First, parse key name.
		devicetype = ValueName[0];
		number = ValueName[1]-48;
		command = ValueName[2];
		switch1 = ValueName[3]-48;
		switch2 = ValueName[4]-48;
		shift = ValueName[5]-48;
		
		// Sanity check.
		DbgPrn(("Sw1: %d    Sw2: %d    Shift: %d\n",switch1,switch2,shift));
		if ((switch1 > 3) || (switch1 < 0 ) || (switch2>3) || (switch2<0) || (shift > 2) || (shift < 0))
		{
			return STATUS_SUCCESS;		
		}
		
	
		if (devExt->Configurations[switch1][switch2][shift] == NULL) 
		{
			devExt->Configurations[switch1][switch2][shift] = ExAllocatePool(NonPagedPool, sizeof(CONFIG));
			if (devExt->Configurations[switch1][switch2][shift] == NULL) 
				return STATUS_INSUFFICIENT_RESOURCES;
			RtlZeroMemory(devExt->Configurations[switch1][switch2][shift], sizeof(CONFIG));
		}
		
		switch (devicetype)
		{
			case 'a': // Axis.
				// Sanity check.
				if ((number > 5) || (number < 0) || (ValueLength != 2)) return STATUS_SUCCESS;
			
				switch (command)
				{
					case 'x': // Maxiumum.
						devExt->Configurations[0][0][0]->Axis[number].Max=*(unsigned short*)ValueData;
						break;
						
					case 'n': // Minimum.
						devExt->Configurations[0][0][0]->Axis[number].Min=*(unsigned short*)ValueData;
						break;
						
					case 'm': // Multiplier.
						break;
						
					case 'u': // Up.
						break;
						
					case 'd': // Down.
						break;
						
					default: // Nothing.
						;
				}
				break;
			
			case 'b': // Button.
				// Sanity check.
				if ((number > 7) || (number < 0)) return STATUS_SUCCESS;
	
				switch (command)
				{
					case 'p': // Pressed.
//						devExt->Configurations[switch1][switch2][shift]->Button[number].Pressed = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Button[number].Pressed = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);
						CurrentCommandNum++;						
						break;
						
					case 'r': // Released.
//						devExt->Configurations[switch1][switch2][shift]->Button[number].Released = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Button[number].Released= CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);
						CurrentCommandNum++;						
						break;
						
					case 'h': // Hold.
//						devExt->Configurations[switch1][switch2][shift]->Button[number].Hold = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Button[number].Hold = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;						
						break;
						
					default: // Nothing.					
						;
				}		
				break;
				
			case 's': // Switch.
				// Sanity check.
				if ((number > 1) || (number < 0)) return STATUS_SUCCESS;
			
			
				switch (command)
				{
					case 'p': // Pressed.
//						devExt->Configurations[switch1][switch2][shift]->Slider[number].Pressed = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Slider[number].Pressed = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;						
						break;
						
					case 'r': // Released.
//						devExt->Configurations[switch1][switch2][shift]->Slider[number].Released = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Slider[number].Released = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;						
						break;
						
					case 'h': // Hold.
//						devExt->Configurations[switch1][switch2][shift]->Slider[number].Hold = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Slider[number].Hold = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;						
						break;
						
					default: // Nothing.					
						;
				}				
				break;
			
			case 'h': // Hat.
				// Sanity check.
				if ((number > 3) || (number < 0)) return STATUS_SUCCESS;
	
				// Get direciton.
				direction=ValueName[6]-48;
				if ((direction > 8) || (direction < 0)) return STATUS_SUCCESS;
				
				switch (command)
				{
					case 'p': // Pressed.
//						devExt->Configurations[switch1][switch2][shift]->Hat[number].Pressed[direction] = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Hat[number].Pressed[direction] =CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;
						break;
						
					case 'r': // Released.
//						devExt->Configurations[switch1][switch2][shift]->Hat[number].Released[direction] = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Hat[number].Released[direction] = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;
						break;
						
					case 'h': // Hold.
//						devExt->Configurations[switch1][switch2][shift]->Hat[number].Hold[direction] = *(unsigned char*) ValueData;
						devExt->Configurations[switch1][switch2][shift]->Hat[number].Hold[direction] = CurrentCommandNum;
						devExt->Commands[CurrentCommandNum]=ParseScript(ValueData, ValueLength);						
						CurrentCommandNum++;
						break;
	
					case 'm': // Mode
						devExt->Configurations[switch1][switch2][shift]->Hat[number].Mode = *(unsigned char*)ValueData; // Parse data somehow?
						
					default: // Nothing.					
						;
				}				
				break;
				
			default:
			
				// Do nothing.
				;
		}
	}
	DbgPrn(("AddEntry complete.\n"));
	return STATUS_SUCCESS;	
}

 

NTSTATUS LoadProfile(struct DEVICE_EXTENSION  *DeviceExtension)
{
	RTL_QUERY_REGISTRY_TABLE myKey[2];
	NTSTATUS status;
	
	DbgPrn(("LoadProfile Begins.\n"));
	
	myKey[0].QueryRoutine=AddEntry;
	myKey[0].Flags=RTL_QUERY_REGISTRY_REQUIRED;
	myKey[0].Name=NULL;
	myKey[0].EntryContext=NULL;
	myKey[0].DefaultType=0;
	myKey[0].DefaultData=NULL;
	myKey[0].DefaultLength=0;
	myKey[1].QueryRoutine=NULL;
	myKey[1].Name=NULL;
	
	CurrentCommandNum=1;	
	status=RtlQueryRegistryValues( RTL_REGISTRY_CONTROL, L"X36", myKey, (PVOID) DeviceExtension, NULL );
	
	DbgPrn(("LoadProfile ends.  Status: %x\n",status));
	return status;
}


void DestroyScript(PSCRIPT Script)
{
	PSCRIPT Current=Script;
	PSCRIPT Next;
	
	while (Current != NULL)
	{
		Next=Current->Next;
		ExFreePool(Current);
		Current=Next;
	}
}
	

NTSTATUS ClearProfile(struct DEVICE_EXTENSION  *DeviceExtension) // Clears a profile and destroys all memory structures.
{
	unsigned int i,j,k,l,m;
	
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			for(k=0;k<3;k++)
			{
				DbgPrn(("Now clearing profile %d %d %d\n",i,j,k));
				if (DeviceExtension->Configurations[i][j][k] != NULL)
				{
					DbgPrn(("Freeing memory\n"));
					ExFreePool(DeviceExtension->Configurations[i][j][k]);
					DeviceExtension->Configurations[i][j][k]=NULL;
				}
			};;;
			
	// Clear scripts.
	for(i=0;i<256;i++)
	{
		if (DeviceExtension->Commands[i] != NULL)
		{
			ExFreePool(DeviceExtension->Commands[i]);
			DeviceExtension->Commands[i]=NULL;
		}
	}


			
	// Init config 0.
//	DbgPrn(("Allocating config 0\n"));
	DeviceExtension->Configurations[0][0][0] = ExAllocatePool(NonPagedPool, sizeof(CONFIG));
	if (DeviceExtension->Configurations[0][0][0] == NULL) 
			return STATUS_INSUFFICIENT_RESOURCES;
//	DbgPrn(("Clearing config 0\n"));			
	RtlZeroMemory(DeviceExtension->Configurations[0][0][0], sizeof(CONFIG));		
	return STATUS_SUCCESS;
}


NTSTATUS ReloadProfile(struct DEVICE_EXTENSION  *DeviceExtension) // Checks to see if a new profile is necessary and reloads if it is.
{
	HANDLE myKey;
	NTSTATUS status;
	ULONG value=0;
	UNICODE_STRING ValueName;
	
    status = IoOpenDeviceRegistryKey(   DeviceExtension->physicalDevObj, 
                                        PLUGPLAY_REGKEY_DEVICE, 
                                        KEY_READ, 
                                        &myKey);

	if (! NT_SUCCESS(status))
		return status;
		
	
	status = QueryDeviceKey( myKey, L"LoadProfile",&value,sizeof(value));
	
	if ((NT_SUCCESS(status))&&(value))
	{		
		DbgPrn(("Reload required.  Reloading profile.\n"));
		// Set reg value to 0.
		value=0;
		
		RtlInitUnicodeString(&ValueName,L"LoadProfile");
		
		ZwSetValueKey( myKey, &ValueName, 0, REG_BINARY, &value, sizeof(&value));
		
		// Clear current profile.
		ClearProfile(DeviceExtension);
		// Load new profile.
		LoadProfile(DeviceExtension);
		
		DbgPrn(("Reload complete.  Reloading profile.\n"));
	}
	ZwClose(myKey);

	DbgPrn(("ReloadProfile ends.  Status: %x\n",status));
	return status;
}
